import datetime
import urllib
import warnings
from abc import abstractmethod
from typing import Union, Any, Dict, List, Optional

from pykechain.enums import FilterType, Category, PropertyType, ScopeStatus
from pykechain.exceptions import IllegalArgumentError, NotFoundError
from pykechain.models.input_checks import (
    check_base,
    check_enum,
    check_type,
    check_text,
    check_datetime,
)
from pykechain.models.widgets.enums import MetaWidget


class BaseFilter:
    """Base class for any filters used in pykechain."""

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__class__.write_options(filters=[self]) == self.__class__.write_options(
                filters=[other]
            )
        else:
            return False

    @classmethod
    @abstractmethod
    def parse_options(cls, options: Dict) -> List["BaseFilter"]:  # pragma: no cover
        """
        Convert the dict & string-based definition of a filter to a list of Filter objects.

        :param options: options dict from a property or meta dict from a widget.
        :return: list of Filter objects
        :rtype list
        """
        pass

    @classmethod
    @abstractmethod
    def write_options(cls, filters: List) -> Dict:
        """
        Convert the list of Filter objects to a dict.

        :param filters: List of BaseFilter objects
        :returns options dict to be used to update the options dict of a property
        """
        if not all(isinstance(f, cls) for f in filters):
            raise IllegalArgumentError(f"All `filters` must be of type `{cls}`")


class PropertyValueFilter(BaseFilter):
    """
    Property value filter, used for Part reference properties and filtered grid widgets.

    :ivar id: property model UUID
    :ivar value: value of the filter
    :ivar type: filter type
    """

    def __init__(
        self,
        property_model: Union[str, "Property"],
        value: Any,
        filter_type: FilterType,
    ):
        """Create PropertyValueFilter instance."""
        from pykechain.models import Property

        property_model_id = check_base(property_model, Property, "property_model")
        check_enum(filter_type, FilterType, "filter_type")

        self.id = property_model_id
        self.value = value
        self.type = filter_type

    def __repr__(self):
        return f"PropertyValueFilter {self.type}: {self.value} ({self.id})"

    def format(self) -> str:
        """Format PropertyValueFilter as a string."""
        if isinstance(self.value, str):
            value = urllib.parse.quote(self.value)
        elif isinstance(self.value, bool):
            value = str(self.value).lower()
        else:
            value = self.value
        return f"{self.id}:{value}:{self.type}"

    def validate(self, part_model: "Part") -> None:
        """
        Validate data of the PropertyValueFilter.

        :param part_model: Part model to which the filter will be applied.
        :returns None
        """
        from pykechain.models import Part

        check_base(part_model, Part, "part_model")
        try:
            prop = part_model.property(self.id)
        except NotFoundError:
            raise IllegalArgumentError(
                "Property value filters can only be set on properties belonging to the selected "
                "Part model."
            )

        if prop.category != Category.MODEL:
            raise IllegalArgumentError(
                'Property value filters can only be set on Property models, received "{}".'.format(
                    prop
                )
            )
        else:
            property_type = prop.type
            if (
                property_type
                in (
                    PropertyType.BOOLEAN_VALUE,
                    PropertyType.REFERENCES_VALUE,
                    PropertyType.ACTIVITY_REFERENCES_VALUE,
                )
                and self.type != FilterType.EXACT
            ):
                warnings.warn(
                    "A PropertyValueFilter on a `{}` property should use "
                    "filter type `{}`, not `{}`".format(
                        property_type, FilterType.EXACT, self.type
                    ),
                    Warning,
                )
            elif (
                property_type
                in (
                    PropertyType.TEXT_VALUE,
                    PropertyType.CHAR_VALUE,
                    PropertyType.LINK_VALUE,
                    PropertyType.SINGLE_SELECT_VALUE,
                    PropertyType.USER_REFERENCES_VALUE,
                    PropertyType.SCOPE_REFERENCES_VALUE,
                )
                and self.type != FilterType.CONTAINS
            ):
                warnings.warn(
                    "A PropertyValueFilter on a `{}` property should use "
                    "filter type `{}`, not `{}`".format(
                        property_type, FilterType.CONTAINS, self.type
                    ),
                    Warning,
                )
            elif property_type in (
                PropertyType.INT_VALUE,
                PropertyType.FLOAT_VALUE,
                PropertyType.DATE_VALUE,
                PropertyType.DATETIME_VALUE,
            ) and self.type not in (FilterType.LOWER_THAN_EQUAL, FilterType.GREATER_THAN_EQUAL):
                warnings.warn(
                    "A PropertyValueFilter on a `{}` property should use "
                    "filter type `{}` or `{}`, not `{}`".format(
                        property_type,
                        FilterType.LOWER_THAN_EQUAL,
                        FilterType.GREATER_THAN_EQUAL,
                        self.type,
                    ),
                    Warning,
                )
            elif (
                property_type in (PropertyType.MULTI_SELECT_VALUE,)
                and self.type != FilterType.CONTAINS_SET
            ):
                warnings.warn(
                    "A PropertyValueFilter on a `{}` property should use "
                    "filter type `{}`, not `{}`".format(
                        property_type, FilterType.CONTAINS_SET, self.type
                    ),
                    Warning,
                )
            else:
                pass

    @classmethod
    def parse_options(cls, options: Dict) -> List["PropertyValueFilter"]:
        """
        Convert the dict & string-based definition of a property value filter to a list of PropertyValueFilter objects.

        :param options: options dict from a multi-reference property or meta dict from a filtered grid widget.
        :return: list of PropertyValueFilter objects
        :rtype list
        """
        check_type(options, dict, "options")

        prefilter_string = options.get(MetaWidget.PREFILTERS, {}).get("property_value")
        prefilter_string_list = prefilter_string.split(",") if prefilter_string else []

        prefilters = list()
        for pf_string in prefilter_string_list:
            prefilter_raw = pf_string.split(":")

            if len(prefilter_raw) == 1:  # FIXME encoding problem KE-chain
                prefilter_raw = pf_string.split("%3A")

            prefilters.append(PropertyValueFilter(*prefilter_raw))

        return prefilters

    @classmethod
    def write_options(cls, filters: List) -> Dict:
        """
        Convert the list of Filter objects to a dict.

        :param filters: List of BaseFilter objects
        :returns options dict to be used to update the options dict of a property
        """
        super().write_options(filters=filters)

        prefilters = {"property_value": ",".join([pf.format() for pf in filters])}
        options = {MetaWidget.PREFILTERS: prefilters}

        return options


class ScopeFilter(BaseFilter):
    """
    Scope filter, used on scope reference properties.

    :ivar tag: string
    """

    # map between KE-chain field and Pykechain attribute, and whether the filter is stored as a list (cs-string)
    MAP = [
        ("name__icontains", "name", False),
        ("status__in", "status", False),
        ("due_date__gte", "due_date_gte", False),
        ("due_date__lte", "due_date_lte", False),
        ("start_date__gte", "start_date_gte", False),
        ("start_date__lte", "start_date_lte", False),
        ("progress__gte", "progress_gte", False),
        ("progress__lte", "progress_lte", False),
        ("tags__contains", "tag", True),
        ("team__in", "team", False),
    ]

    def __init__(
        self,
        tag: Optional[str] = None,
        status: Optional[ScopeStatus] = None,
        name: Optional[str] = None,
        team: Optional[Union[str, "Team"]] = None,
        due_date_gte: Optional[datetime.datetime] = None,
        due_date_lte: Optional[datetime.datetime] = None,
        start_date_gte: Optional[datetime.datetime] = None,
        start_date_lte: Optional[datetime.datetime] = None,
        progress_gte: Optional[float] = None,
        progress_lte: Optional[float] = None,
        **kwargs,
    ):
        """Create a ScopeFilter object."""
        from pykechain.models import Team

        filters = [
            tag,
            status,
            name,
            team,
            due_date_gte,
            due_date_lte,
            start_date_gte,
            start_date_lte,
            progress_gte,
            progress_lte,
        ]
        if sum(p is not None for p in filters) + len(kwargs) != 1:
            raise IllegalArgumentError("Every ScopeFilter object must apply only 1 filter!")

        self.status = check_enum(status, ScopeStatus, "status")
        self.name = check_text(name, "name")
        self.due_date_gte = check_datetime(due_date_gte, "due_date_gte")
        self.due_date_lte = check_datetime(due_date_lte, "due_date_lte")
        self.start_date_gte = check_datetime(start_date_gte, "start_date_gte")
        self.start_date_lte = check_datetime(start_date_lte, "start_date_lte")
        self.progress_gte = check_type(progress_gte, float, "progress_gte")
        self.progress_lte = check_type(progress_lte, float, "progress_lte")
        self.tag = check_text(tag, "tag")
        self.team = check_base(team, Team, "team")
        self.extra_filter: dict = kwargs

    def __repr__(self):
        _repr = "ScopeFilter: "
        if self.name:
            _repr += f"name: `{self.name}`"
        elif self.status:
            _repr += f"status `{self.status}`"
        elif self.due_date_gte:
            _repr += f"due date greater or equal than: `{self.due_date_gte}`"
        elif self.due_date_lte:
            _repr += f"due date lesser or equal than: `{self.due_date_lte}`"
        elif self.start_date_gte:
            _repr += f"start date greater or equal than: `{self.start_date_gte}`"
        elif self.start_date_lte:
            _repr += f"start date lesser or equal than: `{self.start_date_lte}`"
        elif self.progress_gte:
            _repr += f"progress greater or equal than: {self.progress_gte * 100}%"
        elif self.progress_lte:
            _repr += f"progress lesser or equal than: {self.progress_lte * 100}%"
        elif self.tag:
            _repr += f"tag `{self.tag}`"
        elif self.team:
            _repr += f"team: `{self.team}`"
        else:
            _repr += f"{self.extra_filter}"

        return _repr

    @classmethod
    def parse_options(cls, options: Dict) -> List["ScopeFilter"]:
        """
        Convert the dict & string-based definition of a scope filter to a list of ScopeFilter objects.

        :param options: options dict from a scope reference property or meta dict from a scopes widget.
        :return: list of ScopeFilter objects
        :rtype list
        """
        check_type(options, dict, "options")

        filters_dict = options.get(MetaWidget.PREFILTERS, {})
        scope_filters = []

        mapping = {field: (attr, is_list) for field, attr, is_list in cls.MAP}

        for field, value in filters_dict.items():
            if field in mapping:
                attr, is_list = mapping[field]

                try:
                    if is_list:
                        values = value.split(",")
                    else:
                        values = [value]
                except AttributeError:
                    values = value

                for item in values:
                    scope_filters.append(cls(**{attr: item}))
            else:
                scope_filters.append(cls(**{field: value}))

        return scope_filters

    @classmethod
    def write_options(cls, filters: List) -> Dict:
        """
        Convert the list of Filter objects to a dict.

        :param filters: List of BaseFilter objects
        :returns options dict to be used to update the options dict of a property
        """
        super().write_options(filters=filters)

        prefilters = dict()
        options = {MetaWidget.PREFILTERS: prefilters}

        for f in filters:  # type: cls
            found = False
            for field, attr, is_list in cls.MAP:
                filter_value = getattr(f, attr)

                if filter_value is not None:
                    if is_list:
                        # creata a string with commaseparted prefilters, the first item directly and
                        # consequent items with a ,
                        # TODO: refactor to create a list and then join them with a ','
                        if field not in prefilters:
                            prefilters[field] = filter_value
                        else:
                            prefilters[field] += f",{filter_value}"
                    else:
                        prefilters[field] = filter_value

                    found = True
                    break

            if not found:
                prefilters.update(f.extra_filter)

        return options



from pykechain.enums import Category, Multiplicity
from pykechain.exceptions import NotFoundError, MultipleFoundError, IllegalArgumentError
from pykechain.models import PartSet, Part
from pykechain.utils import find
from tests.classes import TestBetamax


class TestPartRetrieve(TestBetamax):
    # 1.8
    def test_get_instances_of_a_model(self):
        wheel_model = self.project.model("Wheel")
        wheel_instances = wheel_model.instances()

        self.assertIsInstance(wheel_instances, PartSet)
        for wheel_instance in wheel_instances:
            self.assertEqual(wheel_instance.category, Category.INSTANCE)
            self.assertEqual(wheel_instance.model().id, wheel_model.id)

    def test_get_instances_of_an_instances_raises_notfound(self):
        wheel_instance = self.project.part("Rear Wheel", category=Category.INSTANCE)
        with self.assertRaises(NotFoundError):
            wheel_instance.instances()

    def test_get_single_instance_of_a_model(self):
        bike_model = self.project.model("Bike")
        bike_instance = bike_model.instance()

        self.assertEqual(bike_instance.category, Category.INSTANCE)

    def test_get_single_instance_of_a_multiplicity_model_raises_multiplefounderror(self):
        wheel_model = self.project.model("Wheel")

        with self.assertRaises(MultipleFoundError):
            wheel_model.instance()

    # test added in 1.12.7
    def test_get_single_instance_of_a_model_without_instances_raises_notfounderror(self):
        catalog = self.project.model(name__startswith="Catalog")
        model_without_instances = self.project.create_model(
            parent=catalog, name="model_without_instances", multiplicity=Multiplicity.ZERO_ONE
        )

        with self.assertRaises(NotFoundError):
            model_without_instances.instance()

        # tearDown
        model_without_instances.delete()

    # test added in 2.1, changed in 3.2
    def test_get_parts_with_descendants_tree(self):
        # setUp
        root = self.project.part(name="Product")
        root.populate_descendants()

        # testing
        self.assertIsInstance(root._cached_children, list)
        self.assertEqual(
            1, len(root._cached_children), msg="Number of instances has changed, expected 1"
        )

        # follow-up
        bike_part = find(root._cached_children, lambda d: d.name == "Bike")
        self.assertIsNotNone(bike_part._cached_children)
        self.assertEqual(
            7,
            len(bike_part._cached_children),
            msg="Number of child instances has changed, expected 7",
        )

    # test added in 2.1, changed in 3.2
    def test_get_models_with_descendants_tree(self):
        # setUp
        root = self.project.model(name="Product")
        root.populate_descendants()

        # testing
        self.assertIsInstance(root._cached_children, list)
        self.assertEqual(
            1, len(root._cached_children), msg="Number of models has changed, expected 1"
        )

        # follow-up
        bike_model = find(root._cached_children, lambda d: d.name == "Bike")
        self.assertIsNotNone(bike_model._cached_children)
        self.assertEqual(
            5,
            len(bike_model._cached_children),
            msg="Number of child models has changed, expected 5",
        )

    # test added in 3.0
    def test_retrieve_parts_with_refs(self):
        # setup
        front_fork_ref = "front-fork"
        front_fork_name = "Front Fork"
        front_fork_part = self.project.part(ref=front_fork_ref)
        front_fork_model = self.project.model(ref=front_fork_ref)

        # testing
        self.assertIsInstance(front_fork_part, Part)
        self.assertEqual(front_fork_name, front_fork_part.name)
        self.assertEqual(Category.INSTANCE, front_fork_part.category)

        self.assertIsInstance(front_fork_model, Part)
        self.assertEqual(front_fork_name, front_fork_model.name)
        self.assertEqual(Category.MODEL, front_fork_model.category)

    def test_child(self):
        root = self.project.model(name="Product")

        bike = root.child(name="Bike")

        self.assertIsInstance(bike, Part)
        self.assertEqual(bike.parent_id, root.id)

        bike_via__call__ = root("Bike")

        self.assertEqual(bike, bike_via__call__)

    def test_child_caching(self):
        root = self.project.model(name="Product")
        self.assertIsNone(root._cached_children, msg="No cached children yet")

        root.children()
        self.assertTrue(root._cached_children, msg="Children should be cached")

        bike = root.child(name="Bike")
        self.assertTrue(root._cached_children, msg="Cache was used and should still be intact")

        bike_again = root.child(pk=bike.id)
        self.assertEqual(bike, bike_again, msg="Bike should be retrieved from cache, based on ID")

        still_the_bike = root.child(name=bike.name)
        self.assertEqual(
            bike, still_the_bike, msg="Bike should be retrieved from cache, based on name"
        )

        root._cached_children = None
        more_bike = root.child(pk=bike.id)
        self.assertEqual(
            bike, more_bike, msg="Cache should be cleared, bike has to be retrieved anew."
        )

    def test_child_invalid(self):
        root = self.project.model(name="Product")

        with self.assertRaises(IllegalArgumentError):
            root.child()

        second_bike_model = root.add_model(name="Bike")
        with self.assertRaises(MultipleFoundError):
            root.child(name="Bike")
        second_bike_model.delete()

        with self.assertRaises(NotFoundError):
            root.child(name="It's only a model")

    def test_all_children(self):
        root = self.project.model(name="Product")

        all_children = root.all_children()

        self.assertIsInstance(all_children, list)
        self.assertEqual(6, len(all_children), msg="Number of models has changed, expected 6")

    def test_child_after_construction(self):
        """Test retrieval of child after creating the model via another Part object of the same KE-chain Part."""
        bike_for_adding = self.project.model(name="Bike")
        bike_for_getting = self.project.model(name="Bike")

        current_children = bike_for_getting.children()

        self.assertEqual(5, len(current_children))

        child = bike_for_adding.add_model(name="__Temp child")
        try:
            retrieved_child = bike_for_getting.child("__Temp child")

            self.assertEqual(child, retrieved_child)
        except NotFoundError as e:
            raise e
        finally:
            # tearDown
            child.delete()



from pykechain.enums import LanguageCodes
from pykechain.exceptions import NotFoundError, MultipleFoundError
from tests.classes import TestBetamax


class TestUsers(TestBetamax):
    def test_retrieve_users(self):
        self.assertTrue(self.client.users())

    def test_retrieve_single_unknown_user(self):
        with self.assertRaises(NotFoundError):
            self.client.user("123lladadwd")

    def test_user_attributes(self):
        attributes = [
            "_client",
            "_json_data",
            "id",
            "name",
            "created_at",
            "updated_at",
            "username",
            "timezone",
            "language",
            "email",
        ]

        obj = self.client.users()[0]
        for attribute in attributes:
            with self.subTest(msg=attribute):
                self.assertTrue(
                    hasattr(obj, attribute),
                    f"Could not find '{attribute}' in the object: '{obj.__dict__}'",
                )

    def test_retrieve_single_multiple_user_raises_error(self):
        with self.assertRaises(MultipleFoundError):
            self.client.user()

    def test_retrieve_single_user_with_known_id(self):
        self.assertTrue(self.client.user(pk=1))

    def test_retrieve_single_user_with_known_username(self):
        self.assertTrue(self.client.user(username="testuser"))

    def test_retrieve_user_timezone(self):
        user_retrieved = self.client.user(username="testuser")
        timezone_retrieved = user_retrieved.timezone
        self.assertIn(timezone_retrieved.zone, ["Atlantic/Cape_Verde", "UTC"])

    def test_retrieve_user_name(self):
        user_retrieved = self.client.user(username="testuser")
        name_retrieved = user_retrieved.name
        self.assertIn(name_retrieved, ["User Test", ""])

    def test_retrieve_default_name(self):
        user_retrieved = self.client.user(username="testuser")
        default_name = user_retrieved.default_name
        self.assertIsInstance(default_name, str)
        self.assertIn(default_name, user_retrieved.username)

    def test_retrieve_user_email(self):
        user_retrieved = self.client.user(username="testuser")
        email_retrieved = user_retrieved.email
        self.assertEqual("a@b.nl", email_retrieved)

    def test_retrieve_user_language(self):
        user_retrieved = self.client.user(username="testuser")
        language_retrieved = user_retrieved.language
        self.assertIn(language_retrieved, LanguageCodes.values())

    def test_now_in_my_timezone(self):
        user_retrieved = self.client.user(username="testuser")
        now = user_retrieved.now_in_my_timezone()
        import datetime

        self.assertIsInstance(now, datetime.datetime)



from ssl import SSLError

from unittest import TestCase

from pykechain.client_utils import PykeRetry
from pykechain.defaults import (
    RETRY_BACKOFF_FACTOR,
    RETRY_ON_CONNECTION_ERRORS,
    RETRY_ON_READ_ERRORS,
    RETRY_ON_REDIRECT_ERRORS,
    RETRY_TOTAL,
)
from urllib3.exceptions import MaxRetryError


class TestPykeRetry(TestCase):
    def test_short_circuit_on_self_signed_cert_error(self):
        """Retry should return early if self signed cert verification fails"""
        retry = PykeRetry(
            total=RETRY_TOTAL,
            connect=RETRY_ON_CONNECTION_ERRORS,
            read=RETRY_ON_READ_ERRORS,
            redirect=RETRY_ON_REDIRECT_ERRORS,
            backoff_factor=RETRY_BACKOFF_FACTOR,
        )

        with self.assertRaises(MaxRetryError):
            retry.increment(
                error=SSLError(
                    "[SSL: CERTIFICATE_VERIFY_FAILED] certificate verify failed: self signed"
                    " certificate"
                )
            )



import datetime

from pykechain.enums import PropertyType
from pykechain.exceptions import IllegalArgumentError
from pykechain.models import DatetimeProperty, Property
from tests.classes import TestBetamax


class TestPropertyDateTime(TestBetamax):
    example_time = datetime.datetime.now()
    NAME = "_TESTING DATE"

    def setUp(self) -> None:
        super().setUp()

        self.bike_model = self.project.model("Bike")
        self.property_model = self.bike_model.add_property(
            name=self.NAME, property_type=PropertyType.DATETIME_VALUE
        )  # type: DatetimeProperty
        self.property = self.project.part("Bike").property(
            name=self.NAME
        )  # type: DatetimeProperty

        self.property.value = self.example_time

    def tearDown(self):
        self.property_model.delete()
        super().tearDown()

    def test_get_datetime(self):
        value = self.property.value

        self.assertIsInstance(value, str)

    def test_set_value_none(self):
        self.property.value = None

    def test_set_value_iso_string(self):
        self.property.value = self.example_time.isoformat()

    def test_set_value_non_datetime(self):
        with self.assertRaises(IllegalArgumentError):
            self.property.value = "3"

    def test_set_value(self):
        self.property.value = self.example_time

    def test_pending_value(self):
        self.property.value = None

        Property.set_bulk_update(True)

        self.property.value = self.example_time

        live_property = self.project.property(id=self.property.id)
        self.assertFalse(live_property.has_value())

        Property.update_values(client=self.client)

        live_property.refresh()
        self.assertTrue(live_property.has_value())

    def test_value_via_part(self):
        self.bike_model.update(
            update_dict={
                self.NAME: self.example_time,
            }
        )

    def test_to_datetime(self):
        date_time_value = self.property.to_datetime()

        self.assertIsInstance(date_time_value, datetime.datetime)

    def test_iso_formatted(self):
        date_time_value = self.property.to_datetime()

        iso_format_datetime = DatetimeProperty.to_iso_format(date_time_value)

        self.assertIsInstance(iso_format_datetime, str)



from datetime import datetime
from typing import Union, Dict, Optional, List  # noqa: F401

import requests

from pykechain.defaults import API_EXTRA_PARAMS
from pykechain.enums import (
    Multiplicity,
    ScopeStatus,
    SubprocessDisplayMode,
    KEChainPages,
    ScopeRoles,
    ScopeMemberActions,
    ScopeCategory,
)
from pykechain.exceptions import APIError, NotFoundError, IllegalArgumentError
from pykechain.models.activity import Activity
from pykechain.models.base import Base
from pykechain.models.context import Context
from pykechain.models.input_checks import (
    check_text,
    check_datetime,
    check_enum,
    check_list_of_text,
    check_base,
    check_type,
)
from pykechain.models.part import Part
from pykechain.models.property import Property
from pykechain.models.representations import BaseRepresentation
from pykechain.models.representations.component import RepresentationsComponent
from pykechain.models.service import Service, ServiceExecution
from pykechain.models.sidebar.sidebar_manager import SideBarManager
from pykechain.models.tags import TagsMixin
from pykechain.models.team import Team
from pykechain.utils import parse_datetime, find, Empty, clean_empty_values, empty


class Scope(Base, TagsMixin):
    """A virtual object representing a KE-chain scope.

    :ivar id: id of the activity
    :type id: uuid
    :ivar name: name of the activity
    :type name: basestring
    :ivar created_at: created datetime of the activity
    :type created_at: datetime
    :ivar updated_at: updated datetime of the activity
    :type updated_at: datetime
    :ivar description: description of the activity
    :type description: basestring
    :ivar workflow_root: uuid of the workflow root object
    :type workflow_root: uuid
    :ivar status: status of the scope. One of :class:`pykechain.enums.ScopeStatus`
    :type status: basestring
    :ivar type: Type of the Scope. One of :class:`pykechain.enums.ScopeType` for WIM version 2
    :type type: basestring
    """

    def __init__(self, json: Dict, **kwargs) -> None:
        """Construct a scope from provided json data."""
        super().__init__(json, **kwargs)

        # for 'kechain2.core.wim <2.0.0'
        self.process = json.get("process")
        # for 'kechain2.core.wim >=2.0.0'
        self.workflow_root = json.get("workflow_root_id")

        self._workflow_root_process = None
        self._catalog_root_process = None
        self._app_root_process = None
        self._product_root_model = None
        self._product_root_instance = None
        self._catalog_root_model = None
        self._catalog_root_instance = None

        self.ref = json.get("ref")
        self.description = json.get("text")
        self.status = json.get("status")
        self.category = json.get("category")

        self._tags = json.get("tags")

        self.start_date = parse_datetime(json.get("start_date"))
        self.due_date = parse_datetime(json.get("due_date"))

        self._representations_container = RepresentationsComponent(
            self,
            self.options.get("representations", {}),
            self._save_representations,
        )

    @property
    def team(self) -> Optional[Team]:
        """Team to which the scope is assigned."""
        team_dict = self._json_data.get("team_id_name")
        if team_dict and team_dict.get("id"):
            return self._client.team(pk=team_dict.get("id"))
        else:
            return None

    @property
    def options(self) -> Dict:
        """Options of the Scope.

        .. versionadded: 3.0
        """
        return self._json_data.get("scope_options")

    @options.setter
    def options(self, option_value):
        self.edit(options=option_value)

    def refresh(self, json=None, url=None, extra_params=None):
        """Refresh the object in place."""
        super().refresh(
            json=json,
            url=self._client._build_url("scope", scope_id=self.id),
            extra_params=API_EXTRA_PARAMS["scope"],
        )

    @property
    def representations(self) -> List["BaseRepresentation"]:
        """Get and set the scope representations."""
        return self._representations_container.get_representations()

    @representations.setter
    def representations(self, value):
        self._representations_container.set_representations(value)

    def _save_representations(self, representation_options):
        options = self.options
        options.update({"representations": representation_options})
        self.options = options

    @property
    def workflow_root_process(self) -> "Activity":
        """Retrieve the Activity root object with classification WORKFLOW."""
        if self._workflow_root_process is None:
            self._workflow_root_process = self.activity(id=self._json_data["workflow_root_id"])
        return self._workflow_root_process

    @property
    def app_root_process(self) -> "Activity":
        """Retrieve the Activity root object with classification APP."""
        if self._app_root_process is None:
            self._app_root_process = self.activity(id=self._json_data["app_root_id"])
        return self._app_root_process

    @property
    def catalog_root_process(self) -> "Activity":
        """Retrieve the Activity root object with classification CATALOG."""
        if self._catalog_root_process is None:
            self._catalog_root_process = self.activity(id=self._json_data["catalog_root_id"])
        return self._catalog_root_process

    @property
    def product_root_model(self) -> "Part":
        """Retrieve the Part root object with classification PRODUCT and category MODEL."""
        if self._product_root_model is None:
            self._product_root_model = self.model(id=self._json_data["product_model_id"])
        return self._product_root_model

    @property
    def product_root_instance(self) -> "Part":
        """Retrieve the Part root object with classification PRODUCT and category INSTANCE."""
        if self._product_root_instance is None:
            self._product_root_instance = self.part(id=self._json_data["product_instance_id"])
        return self._product_root_instance

    @property
    def catalog_root_model(self) -> "Part":
        """Retrieve the Part root object with classification CATALOG and category MODEL."""
        if self._catalog_root_model is None:
            self._catalog_root_model = self.model(id=self._json_data["catalog_model_id"])
        return self._catalog_root_model

    @property
    def catalog_root_instance(self) -> "Part":
        """Retrieve the Part root object with classification CATALOG and category INSTANCE."""
        if self._catalog_root_instance is None:
            self._catalog_root_instance = self.part(id=self._json_data["catalog_instance_id"])
        return self._catalog_root_instance

    #
    # CRUD methods
    #

    def _update_scope_project_team(self, action, role, user):
        """
        Update the Project Team of the Scope. Updates include addition or removing of managers or members.

        :param action: type of action to be applied
        :type action: ScopeMemberActions
        :param role: type of role to be applied to the user
        :type role: ScopeRoles
        :param user: the username of the user to which the action applies to
        :type user: basestring
        :raises APIError: When unable to update the scope project team.
        """
        action = check_enum(action, ScopeMemberActions, "action")
        role = check_enum(role, ScopeRoles, "role")
        user = check_text(user, "user")

        users: List[Dict] = self._client._retrieve_users()["results"]
        user_object: Dict = find(users, lambda u: u["username"] == user)
        if user_object is None:
            raise NotFoundError(f'User "{user}" does not exist')

        url = self._client._build_url(f"scope_{action}_{role}", scope_id=self.id)

        response = self._client._request(
            "PUT",
            url,
            params=API_EXTRA_PARAMS[self.__class__.__name__.lower()],
            data={"user_id": user_object["pk"]},
        )

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError(f"Could not {action} {role} in Scope", response=response)

        self.refresh(json=response.json().get("results")[0])

    def edit(
        self,
        name: Optional[Union[str, Empty]] = empty,
        description: Optional[Union[str, Empty]] = empty,
        start_date: Optional[Union[datetime, Empty]] = empty,
        due_date: Optional[Union[datetime, Empty]] = empty,
        status: Optional[Union[str, ScopeStatus, Empty]] = empty,
        category: Optional[Union[str, ScopeCategory, Empty]] = empty,
        tags: Optional[Union[List[str], Empty]] = empty,
        team: Optional[Union[Team, str, Empty]] = empty,
        options: Optional[Union[Dict, Empty]] = empty,
        **kwargs,
    ) -> None:
        """
        Edit the details of a scope.

        Setting an input to None will clear out the value (exception being name and status).

        :param name: (optionally) edit the name of the scope. Name cannot be cleared.
        :type name: basestring or None or Empty
        :param description: (optionally) edit the description of the scope or clear it
        :type description: basestring or None or Empty
        :param start_date: (optionally) edit the start date of the scope as a datetime object (UTC time/timezone
                            aware preferred) or clear it
        :type start_date: datetime or None or Empty
        :param due_date: (optionally) edit the due_date of the scope as a datetime object (UTC time/timzeone
                            aware preferred) or clear it
        :type due_date: datetime or None or Empty
        :param status: (optionally) edit the status of the scope as a string based. Status cannot be cleared.
        :type status: ScopeStatus or Empty
        :param category (optionally) edit the category of the scope
        :type category: ScopeCategory or Empty
        :param tags: (optionally) replace the tags on a scope, which is a list of strings ["one","two","three"] or
                    clear them
        :type tags: list of basestring or None or Empty
        :param team: (optionally) add the scope to a team. Team cannot be cleared.
        :type team: UUIDstring or None or Empty
        :param options: (optionally) custom options dictionary stored on the scope object
        :type options: dict or None or Empty

        :raises IllegalArgumentError: if the type of the inputs is not correct
        :raises APIError: if another Error occurs
        :warns: UserWarning - When a naive datetime is provided. Defaults to UTC.

        Examples
        --------
        >>> from datetime import datetime
        >>> project.edit(name='New project name',description='Changing the description just because I can',
        ... start_date=datetime.now(),status=ScopeStatus.CLOSED)

        If we want to provide timezone aware datetime objects we can use the 3rd party convenience library :mod:`pytz`.
        Mind that we need to fetch the timezone first and use `<timezone>.localize(<your datetime>)` to make it
        work correctly.

        Using `datetime(2017,6,1,23,59,0 tzinfo=<tz>)` does NOT work for most timezones with a
        daylight saving time. Check the `pytz <http://pythonhosted.org/pytz/#localized-times-and-date-arithmetic>`_
        documentation.

        To make it work using :mod:`pytz` and timezone aware :mod:`datetime` see the following example::

        >>> import pytz
        >>> start_date_tzaware = datetime.now(pytz.utc)
        >>> mytimezone = pytz.timezone('Europe/Amsterdam')
        >>> due_date_tzaware = mytimezone.localize(datetime(2019, 10, 27, 23, 59, 0))
        >>> project.edit(start_date=start_date_tzaware,due_date=due_date_tzaware)

        To assign a scope to a team see the following example::

        >>> my_team = client.team(name='My own team')
        >>> project.edit(team=my_team)

        Not mentioning an input parameter in the function will leave it unchanged. Setting a parameter as None will
        clear its value (where that is possible). The example below will clear the due_date, but leave everything else
        unchanged.

        >>> project.edit(due_date=None)

        """
        update_dict = {
            "id": self.id,
            "name": check_text(name, "name") or self.name,
            "text": check_text(description, "description") or "",
            "start_date": check_datetime(start_date, "start_date"),
            "due_date": check_datetime(due_date, "due_date"),
            "status": check_enum(status, ScopeStatus, "status") or self.status,
            "category": check_enum(category, ScopeCategory, "category"),
            "tags": check_list_of_text(tags, "tags", True) or list(),
            "team_id": check_base(team, Team, "team") or "",
            "scope_options": check_type(options, dict, "options") or dict(),
        }

        if kwargs:  # pragma: no cover
            update_dict.update(kwargs)

        update_dict = clean_empty_values(update_dict=update_dict)

        url = self._client._build_url("scope", scope_id=self.id)

        response = self._client._request(
            "PUT",
            url,
            params=API_EXTRA_PARAMS[self.__class__.__name__.lower()],
            json=update_dict,
        )

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError(f"Could not update Scope {self}", response=response)

        self.refresh(json=response.json().get("results")[0])

        # TODO tags that are set are not in response
        if tags is not None and not isinstance(tags, Empty):
            self._tags = tags

    def clone(self, **kwargs) -> "Scope":
        """Clone a scope.

        See :method:`pykechain.Client.clone_scope()` for available parameters.
        """
        return self._client.clone_scope(source_scope=self, **kwargs)

    def delete(self, asynchronous=True):
        """Delete the scope.

        Only works with enough permissions.

        .. versionadded: 3.0

        See :method:`pykechain.Client.delete_scope()` for available parameters.
        :raises ForbiddenError: if you do not have the permissions to delete a scope
        """
        return self._client.delete_scope(scope=self, asynchronous=asynchronous)

    #
    # Part methods
    #

    def parts(self, *args, **kwargs) -> List["Part"]:
        """Retrieve parts belonging to this scope.

        This uses

        See :class:`pykechain.Client.parts` for available parameters.
        """
        return self._client.parts(*args, scope_id=self.id, **kwargs)

    def part(self, *args, **kwargs) -> "Part":
        """Retrieve a single part belonging to this scope.

        See :class:`pykechain.Client.part` for available parameters.
        """
        return self._client.part(*args, scope_id=self.id, **kwargs)

    def properties(self, *args, **kwargs) -> List["Property"]:
        """Retrieve properties belonging to this scope.

        .. versionadded: 3.0

        See :class:`pykechain.Client.properties` for available parameters.
        """
        return self._client.properties(*args, scope_id=self.id, **kwargs)

    def property(self, *args, **kwargs) -> "Property":
        """Retrieve a single property belonging to this scope.

        .. versionadded: 3.0

        See :class:`pykechain.Client.property` for available parameters.
        """
        return self._client.property(*args, scope_id=self.id, **kwargs)

    def model(self, *args, **kwargs) -> "Part":
        """Retrieve a single model belonging to this scope.

        See :class:`pykechain.Client.model` for available parameters.
        """
        return self._client.model(*args, scope_id=self.id, **kwargs)

    def create_model(self, parent, name, multiplicity=Multiplicity.ZERO_MANY) -> "Part":
        """Create a single part model in this scope.

        See :class:`pykechain.Client.create_model` for available parameters.
        """
        return self._client.create_model(parent, name, multiplicity=multiplicity)

    def create_model_with_properties(
        self, parent, name, multiplicity=Multiplicity.ZERO_MANY, properties_fvalues=None, **kwargs
    ) -> "Part":
        """Create a model with its properties in a single API request.

        See :func:`pykechain.Client.create_model_with_properties()` for available parameters.
        """
        return self._client.create_model_with_properties(
            parent,
            name,
            multiplicity=multiplicity,
            properties_fvalues=properties_fvalues,
            **kwargs,
        )

    #
    # Activity methods
    #

    def activities(self, *args, **kwargs) -> List["Activity"]:
        """Retrieve activities belonging to this scope.

        See :class:`pykechain.Client.activities` for available parameters.
        """
        return self._client.activities(*args, scope=self.id, **kwargs)

    def activity(self, *args, **kwargs) -> "Activity":
        """Retrieve a single activity belonging to this scope.

        See :class:`pykechain.Client.activity` for available parameters.
        """
        return self._client.activity(*args, scope=self.id, **kwargs)

    def create_activity(self, *args, **kwargs) -> "Activity":
        """Create a new activity belonging to this scope.

        See :class:`pykechain.Client.create_activity` for available parameters.
        """
        return self._client.create_activity(self.workflow_root, *args, **kwargs)

    def side_bar(self, *args, **kwargs) -> Optional[SideBarManager]:
        """Retrieve the side-bar manager."""
        return SideBarManager(scope=self, *args, **kwargs)

    def set_landing_page(
        self,
        activity: Union["Activity", KEChainPages],
        task_display_mode: Optional[SubprocessDisplayMode] = SubprocessDisplayMode.ACTIVITIES,
    ) -> None:
        """
        Update the landing page of the scope.

        :param activity: Activity object or KEChainPages option
        :type activity: (Activity, KEChainPages)
        :param task_display_mode: display mode of the activity in KE-chain
        :type task_display_mode: SubprocessDisplayMode
        :return: None
        :rtype None
        """
        from pykechain.models import Activity

        if not (isinstance(activity, Activity) or activity in KEChainPages.values()):
            raise IllegalArgumentError(
                'activity must be of class Activity or a KEChainPages option, "{}" is not.'.format(
                    activity
                )
            )

        check_enum(task_display_mode, SubprocessDisplayMode, "task_display_mode")

        if isinstance(activity, Activity):
            url = f"#/scopes/{self.id}/{task_display_mode}/{activity.id}"
        else:
            url = f"#/scopes/{self.id}/{activity}"

        options = dict(self.options)
        options.update({"landingPage": url})
        self.options = options

    def get_landing_page_url(self) -> Optional[str]:
        """
        Retrieve the landing page URL, if it is set in the options.

        :return: Landing page url
        """
        return self.options.get("landingPage")

    #
    # Service Methods
    #

    def services(self, *args, **kwargs) -> List["Service"]:
        """Retrieve services belonging to this scope.

        See :class:`pykechain.Client.services` for available parameters.

        .. versionadded:: 1.13
        """
        return self._client.services(*args, scope=self.id, **kwargs)

    def create_service(self, *args, **kwargs) -> "Service":
        """Create a service to current scope.

        See :class:`pykechain.Client.create_service` for available parameters.

        .. versionadded:: 1.13
        """
        return self._client.create_service(*args, scope=self.id, **kwargs)

    def service(self, *args, **kwargs) -> "Service":
        """Retrieve a single service belonging to this scope.

        See :class:`pykechain.Client.service` for available parameters.

        .. versionadded:: 1.13
        """
        return self._client.service(*args, scope=self.id, **kwargs)

    def service_executions(self, *args, **kwargs) -> List["ServiceExecution"]:
        """Retrieve services belonging to this scope.

        See :class:`pykechain.Client.service_executions` for available parameters.

        .. versionadded:: 1.13
        """
        return self._client.service_executions(*args, scope=self.id, **kwargs)

    def service_execution(self, *args, **kwargs) -> "ServiceExecution":
        """Retrieve a single service execution belonging to this scope.

        See :class:`pykechain.Client.service_execution` for available parameters.

        .. versionadded:: 1.13
        """
        return self._client.service_execution(*args, scope=self.id, **kwargs)

    #
    # User and Members of the Scope
    #

    def members(
        self,
        is_manager: Optional[bool] = None,
        is_supervisor: Optional[bool] = None,
        is_leadmember: Optional[bool] = None,
    ) -> List[Dict]:
        """
        Retrieve members of the scope.

        .. versionchanged:: 3.7
           we added the supervisor members for backend that support this.

        :param is_manager: (otional) set to True/False to filter members that are/aren't managers, resp.
        :type is_manager: bool
        :param is_supervisor: (optional) set to True/False to filter members that are/aren't supervisors, resp.
        :type is_supervisor: bool
        :param is_leadmember: (optional) set to True/False to filter members that are/aren't leadmembers, resp.
        :type is_leadmember: bool
        :return: List of members, each defined as a dict

        Examples
        --------
        >>> members = project.members()
        >>> managers = project.members(is_manager=True)
        >>> supervisors = project.members(is_supervisor=True)
        >>> leadmembers = project.members(is_leadmember=True)

        """
        members = [member for member in self._json_data["members"] if member["is_active"]]

        if is_manager is not None:
            members = [member for member in members if member.get("is_manager") == is_manager]
        if is_supervisor is not None:
            members = [
                member for member in members if member.get("is_supervisor") == is_supervisor
            ]
        if is_leadmember is not None:
            members = [
                member for member in members if member.get("is_leadmember") == is_leadmember
            ]
        return members

    def add_member(self, member: str) -> None:
        """
        Add a single member to the scope.

        You may only edit the list of members if the pykechain credentials allow this.

        :param member: single username to be added to the scope list of members
        :type member: basestring
        :raises APIError: when unable to update the scope member
        """
        self._update_scope_project_team(
            action=ScopeMemberActions.ADD, role=ScopeRoles.MEMBER, user=member
        )

    def remove_member(self, member: str) -> None:
        """
        Remove a single member to the scope.

        :param member: single username to be removed from the scope list of members
        :type member: basestring
        :raises APIError: when unable to update the scope member
        """
        self._update_scope_project_team(
            action=ScopeMemberActions.REMOVE, role=ScopeRoles.MEMBER, user=member
        )

    def add_manager(self, manager: str) -> None:
        """
        Add a single manager to the scope.

        :param manager: single username to be added to the scope list of managers
        :type manager: basestring
        :raises APIError: when unable to update the scope manager
        """
        self._update_scope_project_team(
            action=ScopeMemberActions.ADD, role=ScopeRoles.MANAGER, user=manager
        )

    def remove_manager(self, manager: str) -> None:
        """
        Remove a single manager to the scope.

        :param manager: single username to be added to the scope list of managers
        :type manager: basestring
        :raises APIError: when unable to update the scope manager
        """
        self._update_scope_project_team(
            action=ScopeMemberActions.REMOVE, role=ScopeRoles.MANAGER, user=manager
        )

    def add_leadmember(self, leadmember: str) -> None:
        """
        Add a single leadmember to the scope.

        :param leadmember: single username to be added to the scope list of leadmembers
        :type leadmember: basestring
        :raises APIError: when unable to update the scope leadmember
        """
        self._update_scope_project_team(
            action=ScopeMemberActions.ADD, role=ScopeRoles.LEADMEMBER, user=leadmember
        )

    def remove_leadmember(self, leadmember: str) -> None:
        """
        Remove a single leadmember to the scope.

        :param leadmember: single username to be added to the scope list of leadmembers
        :type leadmember: basestring
        :raises APIError: when unable to update the scope leadmember
        """
        self._update_scope_project_team(
            action=ScopeMemberActions.REMOVE,
            role=ScopeRoles.LEADMEMBER,
            user=leadmember,
        )

    def add_supervisor(self, supervisor: str) -> None:
        """
        Add a single supervisor to the scope.

        .. versionadded:: 3.7
           requires backend version 3.7 as well.

        :param supervisor: single username to be added to the scope list of supervisors
        :type supervisor: basestring
        :raises APIError: when unable to update the scope supervisor
        """
        if self._client.match_app_version(label="scope", version="<3.6.0"):
            raise NotImplementedError(
                "Adding and removal of supervisor members to a scope not "
                "possible with this backend version"
            )
        self._update_scope_project_team(
            action=ScopeMemberActions.ADD, role=ScopeRoles.SUPERVISOR, user=supervisor
        )

    def remove_supervisor(self, supervisor: str) -> None:
        """
        Remove a single supervisor to the scope.

        .. versionadded:: 3.7
           requires backend version 3.7 as well.

        :param supervisor: single username to be added to the scope list of supervisors
        :type supervisor: basestring
        :raises APIError: when unable to update the scope supervisor
        """
        if self._client.match_app_version(label="scope", version="<3.6.0"):
            raise NotImplementedError(
                "Adding and removal of supervisor members to a scope not "
                "possible with this backend version"
            )
        self._update_scope_project_team(
            action=ScopeMemberActions.REMOVE,
            role=ScopeRoles.SUPERVISOR,
            user=supervisor,
        )

    #
    # Context Methods
    #
    def context(self, *args, **kwargs) -> Context:
        """
        Retrieve a context object in this scope.

        See :class:`pykechain.Client.context` for available parameters.

        .. versionadded:: 3.11

        :return: a Context object
        """
        return self._client.context(*args, scope=self, **kwargs)

    def contexts(self, *args, **kwargs) -> List[Context]:
        """
        Retrieve one or more contexts object in this scope.

        See :class:`pykechain.Client.contexts` for available parameters.

        .. versionadded:: 3.11

        :return: a list of Context objects
        """
        return self._client.contexts(scope=self, **kwargs)

    def create_context(self, *args, **kwargs) -> Context:
        """
        Create a new Context object of a ContextType in a scope.

        See :class:`pykechain.Client.create_context` for available parameters.

        .. versionadded:: 3.11

        :return: a Context object
        """
        return self._client.create_context(scope=self, **kwargs)



from pykechain.exceptions import _DeprecationMixin
from pykechain.models import AttachmentProperty


class AttachmentProperty2(AttachmentProperty, _DeprecationMixin):
    """A virtual object representing a KE-chain attachment property."""

    pass



from typing import List, Optional

from pykechain.defaults import PARTS_BATCH_LIMIT
from pykechain.exceptions import IllegalArgumentError
from pykechain.models import Activity, Scope, user
from pykechain.models.base_reference import _ReferencePropertyInScope, _ReferenceProperty
from pykechain.models.value_filter import ScopeFilter
from pykechain.utils import get_in_chunks


class ActivityReferencesProperty(_ReferencePropertyInScope):
    """A virtual object representing a KE-chain Activity References property.

    .. versionadded:: 3.7
    """

    REFERENCED_CLASS = Activity

    def _retrieve_objects(self, **kwargs) -> List[Activity]:
        """
        Retrieve a list of Activities.

        :param kwargs: optional inputs
        :return: list of Activity2 objects
        """
        activities = []
        for activity_json in self._value:
            activity = Activity(client=self._client, json=activity_json)
            activity.refresh()  # To populate the object with all expected data
            activities.append(activity)
        return activities


class ScopeReferencesProperty(_ReferenceProperty):
    """A virtual object representing a KE-chain Scope References property.

    .. versionadded: 3.9
    """

    REFERENCED_CLASS = Scope

    def _retrieve_objects(self, **kwargs) -> List[Scope]:
        """
        Retrieve a list of Scopes.

        :param kwargs: optional inputs
        :return: list of Scope2 objects
        """
        scope_ids = self._validate_values()

        scopes = []
        if scope_ids:
            scopes = list()
            for chunk in get_in_chunks(scope_ids, PARTS_BATCH_LIMIT):
                scopes.extend(list(self._client.scopes(id__in=",".join(chunk), status=None)))
        return scopes

    def set_prefilters(
        self,
        prefilters: List[ScopeFilter] = None,
        clear: Optional[bool] = False,
    ) -> None:
        """
        Set pre-filters on the scope reference property.

        :param prefilters: list of Scope Filter objects
        :type prefilters: list
        :param clear: whether all existing pre-filters should be cleared. (default = False)
        :type clear: bool

        :return: None
        """
        if prefilters is not None:
            if not isinstance(prefilters, list) or not all(
                isinstance(pf, ScopeFilter) for pf in prefilters
            ):
                raise IllegalArgumentError(
                    f"`prefilters` must be a list of ScopeFilter objects, `{prefilters}` is not."
                )
        else:
            prefilters = []

        if not clear:
            list_of_prefilters = ScopeFilter.parse_options(options=self._options)
        else:
            list_of_prefilters = list()

        list_of_prefilters += prefilters

        # Only update the options if there are any prefilters to be set, or if the original filters have to overwritten
        if list_of_prefilters or clear:
            self._options.update(ScopeFilter.write_options(filters=list_of_prefilters))
            self.edit(options=self._options)

    def get_prefilters(self) -> List[ScopeFilter]:
        """
        Return a list of ScopeFilter objects currently configured on the property.

        :return: list of ScopeFilter objects
        :rtype list
        """
        return ScopeFilter.parse_options(self._options)


class UserReferencesProperty(_ReferenceProperty):
    """A virtual object representing a KE-chain User References property.

    .. versionadded: 3.9
    """

    REFERENCED_CLASS = user.User

    def _validate_values(self) -> List[str]:
        """
        Check if the `_value` attribute has valid content.

        :return list of UUIDs:
        :rtype list
        """
        if not self._value:
            return []

        object_ids = []
        for value in self._value:
            if isinstance(value, dict) and "pk" in value or "id" in value:
                pk = str(value.get("pk", value.get("id")))
                object_ids.append(pk)
            elif isinstance(value, (int, str)):
                object_ids.append(str(value))
            else:  # pragma: no cover
                raise ValueError(f'Value "{value}" must be a dict with field `pk` or a UUID.')
        return object_ids

    def _retrieve_objects(self, **kwargs) -> List[user.User]:
        """
        Retrieve a list of Users.

        :param kwargs: optional inputs
        :return: list of User objects

        """
        user_ids = self._validate_values()

        users = []
        if user_ids:
            users = list()
            for chunk in get_in_chunks(user_ids, PARTS_BATCH_LIMIT):
                users.extend(list(self._client.users(id__in=",".join(chunk))))
        return users

    def value_ids(self) -> Optional[List[int]]:
        """
        Retrieve the referenced object UUIDs only.

        :return: list of UUIDs
        :rtype list
        """
        return [value.get("pk") for value in self._value] if self.has_value() else None



from pykechain.exceptions import IllegalArgumentError, APIError
from pykechain.models import Activity, Part, Property
from pykechain.models.association import Association
from pykechain.models.input_checks import check_list_of_base
from pykechain.utils import is_uuid
from tests.classes import TestBetamax


class TestAssociations(TestBetamax):
    def setUp(self):
        super().setUp()
        self.task = self.project.create_activity(name="widget_test_task")  # type: Activity

        # Exactly 1 model
        self.frame = self.project.part(name="Frame")  # type: Part
        self.frame_model = self.project.model(name="Frame")

        # Zero or more model
        self.wheel_parent = self.project.model(name="Bike")  # type: Property
        self.wheel_model = self.project.model(name="Wheel")

        widgets_manager = self.task.widgets()
        self.form_widget = widgets_manager.add_propertygrid_widget(
            part_instance=self.frame,
            readable_models=self.frame_model.properties[:2],
            writable_models=[],
        )

        self.table_widget = widgets_manager.add_supergrid_widget(
            part_model=self.wheel_model,
            parent_instance=self.wheel_parent,
            readable_models=[],
            writable_models=self.frame_model.properties[:2],
        )

    def tearDown(self):
        self.task.delete()
        super().tearDown()

    def test_retrieve_associations_interface(self):
        limit = 3
        associations = self.client.associations(limit=limit)

        self.assertIsInstance(associations, list)
        self.assertEqual(limit, len(associations))
        self.assertTrue(all(isinstance(a, Association) for a in associations))

    def test_association_attributes(self):
        attributes = ["_client", "_json_data", "id"]

        association = self.client.associations(limit=1)[0]
        for attribute in attributes:
            with self.subTest(msg=attribute):
                self.assertTrue(
                    hasattr(association, attribute),
                    "Could not find '{}' in the association: '{}'".format(
                        attribute, association.__dict__.keys()
                    ),
                )

    def test_retrieve_associations(self):
        for inputs, nr in [
            (dict(widget=self.form_widget), 4),
            (dict(activity=self.task), 8),
            (dict(part=self.frame_model), 14),
            (dict(part=self.frame), 10),
            (dict(property=self.frame_model.properties[0]), 6),
            (dict(property=self.frame.properties[0]), 4),
            (dict(scope=self.project), 48),
        ]:
            with self.subTest(msg=f"{inputs} should be len={nr}"):
                # setUp
                associations = self.client.associations(limit=100, **inputs)

                # testing
                self.assertEqual(nr, len(associations))

    def test_retrieve_association_incorrect_inputs(self):
        for keyword, value in dict(
            widget="not a widget",
            activity="not a task",
            part="not a part",
            scope="not a scope",
            property="not a property",
        ).items():
            with self.subTest(msg=f'{keyword}: "{value}"'):
                with self.assertRaises(IllegalArgumentError):
                    self.client.associations(limit=1, **{keyword: value})

        for limit in [-5, "3"]:
            with self.subTest(f"limit: {type(limit)} {limit}"):
                with self.assertRaises(IllegalArgumentError):
                    self.client.associations(limit=limit)

    def test_update_widget_associations(self):
        original_associations = self.client.associations(widget=self.form_widget)

        self.assertEqual(4, len(original_associations))

        self.form_widget.update_associations(
            readable_models=[],
            writable_models=self.frame_model.properties[2:3],
        )
        associations = self.client.associations(widget=self.form_widget)

        self.assertEqual(6, len(associations))

    def test_update_associations_empty(self):
        self.client.update_widgets_associations(
            widgets=[self.form_widget],
            associations=[([], [])],
        )

    def test_set_associations(self):
        self.form_widget.set_associations(
            readable_models=[],
            writable_models=self.frame_model.properties,
        )

    def test_set_associations_empty(self):
        self.client.set_widgets_associations(
            widgets=[self.form_widget],
            associations=[([], [])],
        )

    def test_clear_associations(self):
        original_associations = self.client.associations(widget=self.form_widget)

        self.assertEqual(4, len(original_associations))

        self.client.clear_widget_associations(widget=self.form_widget)
        associations = self.client.associations(widget=self.form_widget)

        self.assertEqual(0, len(associations))

    def test_remove_associations(self):
        original_associations = self.client.associations(widget=self.form_widget)

        self.assertEqual(4, len(original_associations))

        self.form_widget.remove_associations(
            models=self.frame_model.properties[0:1],
        )
        associations = self.client.associations(widget=self.form_widget)

        self.assertEqual(2, len(associations))

    def test_validate_widgets_input(self):
        with self.assertRaises(IllegalArgumentError, msg="Widget must be a list!"):
            self.client._validate_associations(widgets="Not a list", associations=[])
        with self.assertRaises(IllegalArgumentError, msg="Not every widget is a widget!"):
            self.client._validate_associations(
                widgets=[self.task.widgets()[0], "Not a widget"], associations=[]
            )
        with self.assertRaises(
            IllegalArgumentError, msg="Second set of associations is not of length 2!"
        ):
            self.client._validate_associations(
                widgets=list(self.task.widgets()), associations=[([], []), ()]
            )
        with self.assertRaises(IllegalArgumentError, msg="Inputs are not of equal length!"):
            self.client._validate_associations(
                widgets=list(self.task.widgets()),
                associations=[([], [])],
            )

    # noinspection PyTypeChecker
    def test_validate_model_input(self):
        model_ids = check_list_of_base(self.frame_model.properties, Property)

        self.assertIsInstance(model_ids, list)
        self.assertTrue(all(is_uuid(model_id) for model_id in model_ids))
        self.assertEqual(len(self.frame_model.properties), len(model_ids))

        with self.assertRaises(IllegalArgumentError):
            check_list_of_base(objects="not a list")

        with self.assertRaises(IllegalArgumentError):
            check_list_of_base(
                objects=[self.frame_model.properties[0], "Not a property"], cls=Property
            )

    def test_validate_widget_input(self):
        for method in [
            self.client.update_widget_associations,
            self.client.set_widget_associations,
            self.client.clear_widget_associations,
            self.client.remove_widget_associations,
        ]:
            with self.assertRaises(IllegalArgumentError):
                method(widget="Not a widget")

    def test_readable_models(self):
        self.client.update_widget_associations(
            widget=self.form_widget,
            readable_models=self.frame_model.properties,
        )

        self.client.update_widget_associations(
            widget=self.table_widget,
            readable_models=self.wheel_model.properties,
        )

        with self.assertRaises(APIError):
            self.client.update_widget_associations(
                widget=self.table_widget,
                readable_models=self.frame.properties,
            )

    def test_writable_models(self):
        self.client.update_widget_associations(
            widget=self.form_widget,
            writable_models=self.frame_model.properties,
        )

        self.client.update_widget_associations(
            widget=self.table_widget,
            writable_models=self.wheel_model.properties,
        )

        with self.assertRaises(APIError):
            self.client.update_widget_associations(
                widget=self.table_widget,
                writable_models=self.frame.properties,
            )



import requests

from pykechain.enums import NotificationStatus, NotificationEvent, NotificationChannels
from pykechain.exceptions import IllegalArgumentError, APIError
from pykechain.models import Base
from typing import List, Optional, Dict, Union

from pykechain.models.input_checks import check_text, check_enum, check_base, check_user
from pykechain.utils import is_valid_email, Empty, clean_empty_values, empty


class Notification(Base):
    """A virtual object representing a KE-chain notification.

    :ivar id: UUID of the notification
    :type id: basestring
    :ivar subject: subject of the notification
    :type subject: basestring or None
    :ivar created_at: the datetime when the object was created if available (otherwise None)
    :type created_at: datetime or None
    :ivar updated_at: the datetime when the object was last updated if available (otherwise None)
    :type updated_at: datetime or None
    :ivar status: The status of the notification (see :class:`pykechain.enums.NotificationStatus`)
    :type status: basestring or None
    :ivar event: The event of the notification (see :class:`pykechain.enums.NotificationEvent`)
    :type event: basestring or None
    :ivar recipient_users: The list of ids of the users
    :type recipient_users: List[User ids]
    """

    def __init__(self, json: Dict, **kwargs) -> None:
        """Construct a notification from a KE-chain 2 json response.

        :param json: the json response to construct the :class:`Notification` from
        :type json: dict
        """
        super().__init__(json, **kwargs)

        self.message = json.get("message", "")
        self.subject: str = json.get("subject", "")
        self.status: str = json.get("status", "")
        self.event: str = json.get("event", "")
        self.channels: List = json.get("channels", list())
        self.recipient_user_ids: List = json.get("recipient_users", list())
        self.team_id: str = json.get("team", "")
        self.from_user_id: str = json.get("from_user", "")

        self._from_user: Optional["User"] = None
        self._recipient_users: Optional[List["User"]] = None
        self._team: Optional["Team"] = None

    def __repr__(self):  # pragma: no cover
        return f"<pyke Notification id {self.id[-8:]}>"

    def get_recipient_users(self) -> List["User"]:
        """Return the list of actual `User` objects based on recipient_users_ids."""
        if self._recipient_users is None:
            self._recipient_users = self._client.users(
                id__in=",".join([str(pk) for pk in self.recipient_user_ids])
            )
        return self._recipient_users

    def get_from_user(self) -> "User":
        """Return the actual `User` object based on the from_user_id."""
        if self._from_user is None and self.from_user_id:
            self._from_user = self._client.user(pk=self.from_user_id)
        return self._from_user

    def get_team(self) -> "Team":
        """Return the actual `Team` object based on the team_id."""
        if self._team is None and self.team_id:
            self._team = self._client.team(pk=self.team_id)
        return self._team

    def delete(self):
        """Delete the notification."""
        return self._client.delete_notification(notification=self)

    def edit(
        self,
        subject: Optional[Union[str, Empty]] = empty,
        message: Optional[Union[str, Empty]] = empty,
        status: Optional[Union[NotificationStatus, Empty]] = empty,
        recipients: Optional[Union[List[Union["User", str, int]], Empty]] = empty,
        team: Optional[Union["Team", str, Empty]] = empty,
        from_user: Optional[Union["User", str, Empty]] = empty,
        event: Optional[Union[NotificationEvent, Empty]] = empty,
        channel: Optional[Union[NotificationChannels, Empty]] = empty,
        **kwargs,
    ) -> None:
        """
        Update the current `Notification` attributes.

        Setting an input to None will clear out the value (only applicable to recipients and from_user).

        :param subject: (O) Header text of the notification. Cannot be cleared.
        :type subject: basestring or None or Empty
        :param message: (O) Content message of the notification. Cannot be cleared.
        :type message: basestring or None or Empty
        :param status: (O) life-cycle status of the notification, defaults to "DRAFT". Cannot be cleared.
        :type status: NotificationStatus
        :param recipients: (O) list of recipients, each being a User object, user ID or an email address.
        :type recipients: list or None or Empty
        :param team: (O) team object to which the notification is constrained
        :type team: Team object or Team UUID
        :param from_user: (O) Sender of the notification, either a User object or user ID. Defaults to script user.
        :type from_user: User or user ID or None or Empty
        :param event: (O) originating event of the notification. Cannot be cleared.
        :type event: NotificationEvent
        :param channel: (O) method used to send the notification, defaults to "EMAIL". Cannot be cleared.
        :type channel: NotificationChannels
        :param kwargs: (optional) keyword=value arguments
        :return: None
        :raises: APIError: when the `Notification` could not be updated

        Not mentioning an input parameter in the function will leave it unchanged. Setting a parameter as None will
        clear its value (where that is possible). The example below will clear the from_user, but leave everything
        else unchanged.

        >>> notification.edit(from_user=None)

        """
        from pykechain.models import Team, User

        recipient_users = list()
        recipient_emails = list()

        if recipients is not None:
            if isinstance(recipients, list) and all(
                isinstance(r, (str, int, User)) for r in recipients
            ):
                for recipient in recipients:
                    if is_valid_email(recipient):
                        recipient_emails.append(recipient)
                    else:
                        recipient_users.append(check_user(recipient, User, "recipient"))
            elif isinstance(recipients, Empty):
                recipient_emails = empty
                recipient_users = empty
            else:
                raise IllegalArgumentError(
                    "`recipients` must be a list of User objects, IDs or email addresses, "
                    '"{}" ({}) is not.'.format(recipients, type(recipients))
                )

        if isinstance(channel, Empty):
            channels = empty
        elif check_enum(channel, NotificationChannels, "channel"):
            channels = [channel]
        else:
            channels = list()

        update_dict = {
            "status": check_enum(status, NotificationStatus, "status") or self.status,
            "event": check_enum(event, NotificationEvent, "event") or self.event,
            "subject": check_text(subject, "subject") or self.subject,
            "message": check_text(message, "message") or self.message,
            "recipient_users": recipient_users,
            "recipient_emails": recipient_emails,
            "team": check_base(team, Team, "team"),
            "from_user": check_user(from_user, User, "from_user"),
            "channels": channels or self.channels,
        }

        if kwargs:
            update_dict.update(kwargs)

        update_dict = clean_empty_values(update_dict=update_dict)

        url = self._client._build_url("notification", notification_id=self.id)

        response = self._client._request("PUT", url, json=update_dict)

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError(f"Could not update Notification {self}", response=response)

        self.refresh(json=response.json().get("results")[0])



from abc import ABC, abstractmethod
from typing import Union
from unittest import TestCase

from jsonschema import validate

from pykechain.enums import (
    FontAwesomeMode,
    GeoCoordinateConfig,
    LinkTargets,
    Multiplicity,
    PropertyType,
    SelectListRepresentations,
)
from pykechain.exceptions import APIError, IllegalArgumentError
from pykechain.models import Activity, AnyProperty, Property, Scope, SelectListProperty
from pykechain.models.representations.representation_base import BaseRepresentation
from pykechain.models.representations.representations import (
    Autofill,
    ButtonRepresentation,
    CameraScannerInputRepresentation,
    CustomIconRepresentation,
    DecimalPlaces,
    GeoCoordinateRepresentation,
    LinkTarget,
    SignificantDigits,
    ThousandsSeparator,
    UsePropertyNameRepresentation,
)
from pykechain.models.validators.validator_schemas import options_json_schema
from tests.classes import TestBetamax

representation_config = dict(
    rtype="buttonRepresentation",
    config=dict(
        buttonRepresentation="dropdown",
    ),
)
options = dict(
    representations=[representation_config],
)


class TestRepresentationJSON(TestCase):
    def test_valid_button_representation_json(self):
        validate(options_json_schema, options)


class TestRepresentation(TestCase):
    def test_create(self):
        representation = BaseRepresentation()

        self.assertIsInstance(representation, BaseRepresentation)

    def test_create_with_value(self):
        representation = DecimalPlaces(value=3)

        self.assertEqual(3, representation.value)

    def test_create_with_object(self):
        empty_prop = Property(json={}, client=None)
        representation = ButtonRepresentation(empty_prop)

        self.assertIsInstance(representation, ButtonRepresentation)

    def test_parse(self):
        empty_slp = SelectListProperty(json={"value_options": {}}, client=None)
        BaseRepresentation.parse(obj=empty_slp, json=representation_config)

    def test_parse_incorrect_rtype(self):
        no_rtype_config = dict(
            config=dict(
                buttonRepresentation="dropdown",
            )
        )
        with self.assertRaises(ValueError):
            BaseRepresentation.parse(obj=None, json=no_rtype_config)

    def test_parse_unknown_rtype(self):
        no_rtype_config = dict(
            rtype="Not a representation type",
            config=dict(
                buttonRepresentation="dropdown",
            ),
        )
        with self.assertRaises(TypeError):
            BaseRepresentation.parse(obj=None, json=no_rtype_config)

    def test_component_invalid_object(self):
        empty_activity = Activity(json={"id": "1234567890"}, client=None)
        representation = ThousandsSeparator(empty_activity)

        with self.assertRaises(IllegalArgumentError):
            empty_activity.representations = [representation]

    def test_component_invalid_property_type(self):
        empty_prop = Property(json={"id": "1234567890", "category": "MODEL"}, client=None)
        representation = ThousandsSeparator(empty_prop)
        representation.rtype = "Broken rtype"

        with self.assertRaises(IllegalArgumentError):
            empty_prop.representations = [representation]

    def test_component_not_a_list(self):
        empty_activity = Activity(json={}, client=None)

        with self.assertRaises(IllegalArgumentError):
            empty_activity.representations = "Howdy!"

    def test_component_not_a_representation(self):
        empty_activity = Activity(json={}, client=None)

        with self.assertRaises(IllegalArgumentError):
            empty_activity.representations = ["Howdy again!"]


class Bases:
    """
    Wrapping private test classes in this Base class allows some inheritance mechanisms to be used.

    If the _Test...() classes would not be wrapped, the test framework would try to run these abstract classes.
    """

    class _TestRepresentationLive(TestBetamax):
        """
        Specify these class attributes in every sub-class
        """

        representation_class = None  # type: type(BaseRepresentation)
        value = None
        new_value = None
        incorrect_value = None

        test_object_name = "__test object for representations"

        def setUp(self):
            super().setUp()
            self.obj = self._get_object()
            try:
                self.obj.representations = [
                    self.representation_class(
                        obj=self.obj,
                        value=self.value,
                    ),
                ]
            except Exception as e:
                self.obj.part.delete()
                raise e

        def tearDown(self):
            if self.obj:
                try:
                    self.obj.delete()
                except APIError:
                    pass
            super().tearDown()

        @abstractmethod
        def _get_object(self) -> Union[Property, Scope, Activity]:
            pass

        def test_create_with_prop(self):
            representation = self.representation_class(prop=self.obj, value=self.value)

            self.assertIsInstance(representation, self.representation_class)

        def test_get_set(self):
            """Test representation accessor property of the Mixin class"""
            reloaded_obj = self.client.reload(self.obj)
            self.assertEqual(self.obj, reloaded_obj)

            # setUp: get representations
            first_representation = reloaded_obj.representations[0]

            self.assertIsInstance(first_representation, self.representation_class)
            self.assertEqual(self.value, first_representation.value)

        def test_set_value(self):
            """Test value accessor property of the Representation class"""
            # Set new value via the representation object
            first_repr = self.obj.representations[0]
            first_repr.value = self.new_value

            # Retrieve representation again
            reloaded_obj = self.client.reload(self.obj)
            self.assertEqual(self.obj, reloaded_obj)

            first_representation = reloaded_obj.representations[0]

            # testing
            self.assertEqual(first_repr.value, first_representation.value)

        def test_unsupported_value(self):

            with self.assertRaises(IllegalArgumentError):
                self.representation_class(
                    obj=self.obj,
                    value=self.incorrect_value,
                )

    class _TestPropertyRepresentation(_TestRepresentationLive):
        property_type = None
        test_model_name = "__test model for representations"
        test_model = None
        incorrect_value = "Unsupported value"

        def tearDown(self):
            if self.test_model:
                try:
                    self.test_model.delete()
                except APIError:
                    pass
            super().tearDown()

        def _get_object(self):
            parent_model = self.project.model(name="Bike")
            self.test_model = parent_model.add_model(
                name=self.test_model_name, multiplicity=Multiplicity.ONE
            )
            obj = self.test_model.add_property(
                name=self.test_object_name,
                property_type=self.property_type,
            )  # type: AnyProperty
            return obj

    class _TestCustomIconRepresentation(_TestRepresentationLive, ABC):
        representation_class = CustomIconRepresentation
        value = "university"
        new_value = "bat"
        incorrect_value = ["must be a string"]

        def test_set_mode(self):
            representation = self.obj.representations[0]  # type: CustomIconRepresentation

            # Set mode
            representation.display_mode = FontAwesomeMode.SOLID

            reloaded_obj = self.client.reload(obj=self.obj)
            reloaded_repr = reloaded_obj.representations[0]

            # Test get of mode
            self.assertEqual(representation.display_mode, reloaded_repr.display_mode)

        def test_set_mode_incorrect(self):
            representation = self.obj.representations[0]  # type: CustomIconRepresentation

            with self.assertRaises(IllegalArgumentError):
                representation.display_mode = "fancy colors"


class TestReprSignificantDigits(Bases._TestPropertyRepresentation):
    property_type = PropertyType.FLOAT_VALUE
    representation_class = SignificantDigits
    value = 3
    new_value = 1


class TestReprThousandsSeparator(Bases._TestPropertyRepresentation):
    property_type = PropertyType.FLOAT_VALUE
    representation_class = ThousandsSeparator
    value = None
    new_value = None


class TestReprDecimalPlaces(Bases._TestPropertyRepresentation):
    property_type = PropertyType.FLOAT_VALUE
    representation_class = DecimalPlaces
    value = 3
    new_value = 5


class TestReprLinkTarget(Bases._TestPropertyRepresentation):
    property_type = PropertyType.LINK_VALUE
    representation_class = LinkTarget
    value = LinkTargets.SAME_TAB
    new_value = LinkTargets.NEW_TAB


class TestGeocoordinateRepresentation(Bases._TestPropertyRepresentation):
    property_type = PropertyType.GEOJSON_VALUE
    representation_class = GeoCoordinateRepresentation
    value = GeoCoordinateConfig.RD_AMERSFOORT
    new_value = GeoCoordinateConfig.APPROX_ADDRESS


class TestReprButton(Bases._TestPropertyRepresentation):
    property_type = PropertyType.SINGLE_SELECT_VALUE
    representation_class = ButtonRepresentation
    value = SelectListRepresentations.CHECK_BOXES
    new_value = SelectListRepresentations.BUTTONS

    def setUp(self):
        super().setUp()
        self.obj.options = ["alpha", "beta", "gamma", "omega"]


class TestUsePropertyNameRepresentationForPartReferences(Bases._TestPropertyRepresentation):
    property_type = PropertyType.REFERENCES_VALUE
    representation_class = UsePropertyNameRepresentation
    value = True
    new_value = False

    def test_empty_config_value(self):
        # get the representation on the object
        repr = self.obj.representations[-1]

        # set it as an empty config object
        repr_json = repr.as_json()
        repr_json["config"] = {}
        new_value_options = dict(representations=[repr_json])
        # edit options directly otherwise it would not 'stick'
        self.obj.edit(options=new_value_options)

        # reload from server the representations
        self.obj.refresh()
        repr = self.obj.representations[-1]

        # check sanity of the representation with the empty object.
        self.assertDictEqual(self.obj._options.get("representations")[0], repr_json)
        self.assertEqual(type(repr), UsePropertyNameRepresentation)
        self.assertIsNone(repr.value)


class TestUsePropertyNameRepresentationForUserReferences(Bases._TestPropertyRepresentation):
    property_type = PropertyType.USER_REFERENCES_VALUE
    representation_class = UsePropertyNameRepresentation
    value = True
    new_value = False


class TestUsePropertyNameRepresentationForScopeReferences(Bases._TestPropertyRepresentation):
    property_type = PropertyType.SCOPE_REFERENCES_VALUE
    representation_class = UsePropertyNameRepresentation
    value = True
    new_value = False


class TestUsePropertyNameRepresentationForServiceReferences(Bases._TestPropertyRepresentation):
    property_type = PropertyType.SERVICE_REFERENCES_VALUE
    representation_class = UsePropertyNameRepresentation
    value = True
    new_value = False


class TestUsePropertyNameRepresentationForActivityReferences(Bases._TestPropertyRepresentation):
    property_type = PropertyType.ACTIVITY_REFERENCES_VALUE
    representation_class = UsePropertyNameRepresentation
    value = True
    new_value = False


class TestUseCameraScannerInputRepresentationForCharProperties(Bases._TestPropertyRepresentation):
    property_type = PropertyType.CHAR_VALUE
    representation_class = CameraScannerInputRepresentation
    value = True
    new_value = False


class TestUseCameraScannerInputRepresentationForTextAreaCharProperties(
    Bases._TestPropertyRepresentation
):
    property_type = PropertyType.TEXT_VALUE
    representation_class = CameraScannerInputRepresentation
    value = True
    new_value = False


class TestUseCameraScannerInputRepresentationForFloatAreaProperties(
    Bases._TestPropertyRepresentation
):
    property_type = PropertyType.FLOAT_VALUE
    representation_class = CameraScannerInputRepresentation
    value = True
    new_value = False


class TestUseCameraScannerInputRepresentationForIntegerAreaProperties(
    Bases._TestPropertyRepresentation
):
    property_type = PropertyType.INT_VALUE
    representation_class = CameraScannerInputRepresentation
    value = True
    new_value = False


class TestReprAutofill(Bases._TestPropertyRepresentation):
    property_type = PropertyType.DATETIME_VALUE
    representation_class = Autofill
    value = False
    new_value = True


class TestReprAutofillUser(Bases._TestPropertyRepresentation):
    property_type = PropertyType.USER_REFERENCES_VALUE
    representation_class = Autofill
    value = False
    new_value = True


class TestReprActivity(Bases._TestCustomIconRepresentation):
    def _get_object(self):
        return self.project.create_activity(name=self.test_object_name)


class TestReprScope(Bases._TestCustomIconRepresentation):
    def _get_object(self):
        return self.client.create_scope(name=self.test_object_name)
